/*
Copyright (c) 2014, Romain Dolbeau, unless otherwise noted.
No claims is made upon work by others.

For the work by Romain Dolbeau:
All rights reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

For the work by others:
See the indicated reference for the relevant license.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <malloc.h>

#include "chacha.h"

/* chacha20 implementation based on code by D. J. Bernstein */

#define U32C(v) (v##U)
#if defined(__BIG_ENDIAN__)
#define U32TO32_LITTLE(v) __builtin_bswap32(v)
#define U8TO32_LITTLE(p) U32TO32_LITTLE(((uint32_t*)(p))[0])
#define U32TO8_LITTLE(p, v) (((uint32_t*)(p))[0] = U32TO32_LITTLE(v))
#else
#define U32TO32_LITTLE(v) (v)
#define U8TO32_LITTLE(p) U32TO32_LITTLE(((uint32_t*)(p))[0])
#define U32TO8_LITTLE(p, v) (((uint32_t*)(p))[0] = U32TO32_LITTLE(v))
#endif

#define U32V(v) ((uint32_t)(v) & U32C(0xFFFFFFFF))
#define ROTL32(v, n) \
  (U32V((v) << (n)) | ((v) >> (32 - (n))))

#define ROTATE(v,c) (ROTL32(v,c))
#define XOR(v,w) ((v) ^ (w))
#define PLUS(v,w) (U32V((v) + (w)))
#define PLUSONE(v) (PLUS((v),1))

#define QUARTERROUND(a,b,c,d) \
  x[a] = PLUS(x[a],x[b]); x[d] = ROTATE(XOR(x[d],x[a]),16); \
  x[c] = PLUS(x[c],x[d]); x[b] = ROTATE(XOR(x[b],x[c]),12); \
  x[a] = PLUS(x[a],x[b]); x[d] = ROTATE(XOR(x[d],x[a]), 8); \
  x[c] = PLUS(x[c],x[d]); x[b] = ROTATE(XOR(x[b],x[c]), 7);

void salsa20_wordtobyte(uint8_t output[64],const uint32_t input[16])
{
  uint32_t x[16];
  int i;

  for (i = 0;i < 16;++i) x[i] = input[i];
  for (i = 20;i > 0;i -= 2) {
    QUARTERROUND( 0, 4, 8,12)
    QUARTERROUND( 1, 5, 9,13)
    QUARTERROUND( 2, 6,10,14)
    QUARTERROUND( 3, 7,11,15)
    QUARTERROUND( 0, 5,10,15)
    QUARTERROUND( 1, 6,11,12)
    QUARTERROUND( 2, 7, 8,13)
    QUARTERROUND( 3, 4, 9,14)
  }
  for (i = 0;i < 16;++i) x[i] = PLUS(x[i],input[i]);
  for (i = 0;i < 16;++i) U32TO8_LITTLE(output + 4 * i,x[i]);
}

#ifdef __ARM_NEON__
#include <arm_neon.h>
#endif
#ifdef __x86_64__
#include <immintrin.h>
#endif
#ifdef __powerpc__
#if defined(__ALTIVEC__)
#include <altivec.h>
#endif
#endif
void encrypt_bytes(uint32_t *x, const uint8_t *m, uint8_t *c_, uint32_t bytes)
{
  uint8_t output[64];
  int i;
  uint8_t* out = c_;

#ifdef __ARM_NEON__
#if 1
#include "u4_neon.h"
#endif
#if 0
#include "u1_neon.h"
#endif
#endif

#ifdef __x86_64__
#if defined(__AVX512F__)
#include "u16_x86_64.h"
#endif
#if defined(__AVX2__)
#include "u8_x86_64.h"
#endif
#if 1
#include "u4_x86_64.h"
#endif
#if 1
#include "u1_x86_64.h"
#endif
#endif

#ifdef __MIPSEL__
#if defined(__mips_msa)
#include "u4_msa.h"
#endif
#endif

#ifdef __powerpc__
#if defined(__ALTIVEC__)
#include "u4_altivec.h"
#endif
#endif

  if (!bytes) return;
  for (;;) {
    salsa20_wordtobyte(output,x);
    x[12] = PLUSONE(x[12]);
    if (!x[12]) {
      x[13] = PLUSONE(x[13]);
      /* stopping at 2^70 bytes per nonce is user's responsibility */
    }
    if (bytes <= 64) {
      for (i = 0;i < bytes;++i) out[i] = m[i] ^ output[i];
      return;
    }
    for (i = 0;i < 64;++i) out[i] = m[i] ^ output[i];
    bytes -= 64;
    out += 64;
    m += 64;
  }
}


static const char sigma[16] = "expand 32-byte k";
static const char tau[16] = "expand 16-byte k";
void keysetup(uint32_t *input, const uint8_t *k, const uint32_t kbits, const uint32_t ivbits)
{
  const char *constants;

  input[4] = U8TO32_LITTLE(k + 0);
  input[5] = U8TO32_LITTLE(k + 4);
  input[6] = U8TO32_LITTLE(k + 8);
  input[7] = U8TO32_LITTLE(k + 12);
  if (kbits == 256) { /* recommended */
    k += 16;
    constants = sigma;
  } else { /* kbits == 128 */
    constants = tau;
  }
  input[8] = U8TO32_LITTLE(k + 0);
  input[9] = U8TO32_LITTLE(k + 4);
  input[10] = U8TO32_LITTLE(k + 8);
  input[11] = U8TO32_LITTLE(k + 12);
  input[0] = U8TO32_LITTLE(constants + 0);
  input[1] = U8TO32_LITTLE(constants + 4);
  input[2] = U8TO32_LITTLE(constants + 8);
  input[3] = U8TO32_LITTLE(constants + 12);
}

void ivsetup(uint32_t *input, const uint8_t *iv)
{
  input[12] = 0;
  input[13] = 0;
  input[14] = U8TO32_LITTLE(iv + 0);
  input[15] = U8TO32_LITTLE(iv + 4);
}


#ifdef TEST_CHACHA_CPU

#include <openssl/md5.h>

#define SIZE (16*1024*1024)
int main(int argc, char **argv) {
  uint8_t *in, *out0, *out1;
  uint32_t key[8] = {0xc47b0294, 0xdbbbee0f, 0xec4757f2, 0x2ffeee35, 0x87ca4730, 0xc3d33b69, 0x1df38bab, 0x076bc558 };
  uint32_t iv[2] = {0xdeadbeef, 0xbadc0ffe};
  uint32_t keyinput[16];
  uint32_t i, j;
  double t0, t1, t0b, t1b;
  uint32_t count;

  in   = (uint8_t*)memalign(64, (size_t)SIZE);
  out0 = (uint8_t*)memalign(64, (size_t)SIZE);
  out1 = (uint8_t*)memalign(64, (size_t)SIZE);

  srandom(100);
  for (i = 0 ; i < SIZE/sizeof(uint32_t) ; i++) {
    ((uint32_t*)in)[i] = 0;//U32TO32_LITTLE(random());
  }


#if defined(__BIG_ENDIAN__)
  for (i = 0 ; i < 8 ; i++)
    key[i] = __builtin_bswap32(key[i]);
  for (i = 0 ; i < 2 ; i++)
    iv[i] = __builtin_bswap32(iv[i]);
#endif

  keysetup(keyinput, (const unsigned char *)key, 256, 0);

#define ITER 3
  for (i = 0 ; i < ITER ;i++) {
    ivsetup(keyinput, (const unsigned char *)iv);
    t0 = wallclock();
    encrypt_bytes(keyinput, in, out0, SIZE);
    t1 = wallclock();
    printf("%zd B in %lf s = %lf MB/s\n",
           SIZE, t1-t0, ((double)SIZE)/(1000000. * (t1-t0)));
    {
      unsigned char md5[16];
      MD5(out0,SIZE,md5);
      int i;
      for (i = 0 ; i < 16 ; i++) {
        printf("%02x", md5[i]);
      }
      printf("\n");
/*       for (i = 0 ; i < 256 ; i++) { */
/*         printf("%02x ", out0[i]); */
/*       } */
/*       printf("\n"); */
    }
  }
  free(in);
  free(out0);
  free(out1);
}
#endif
